@@ -2,18 +2,18 @@
 Tool to extract and pretty-print PDF annotations for reviewing.
 """
 
-__version__ = '0.2'
+__version__ = '0.4'
 
 import bisect
 import collections
 import itertools
 import logging
-import typing
+import typing as typ
 
 from pdfminer.pdfinterp import PDFResourceManager, PDFPageInterpreter
 from pdfminer.pdfpage import PDFPage
-from pdfminer.layout import (
-    LAParams, LTContainer, LTAnno, LTChar, LTPage, LTTextBox, LTTextLine, LTItem, LTComponent)
+from pdfminer.layout import (LAParams, LTAnno, LTChar, LTComponent, LTContainer, LTFigure, LTItem,
+                             LTPage, LTTextBox, LTTextLine)
 from pdfminer.converter import PDFLayoutAnalyzer
 from pdfminer.pdfparser import PDFParser
 from pdfminer.pdfdocument import PDFDocument, PDFNoOutlines
@@ -23,20 +23,29 @@ import pdfminer.settings
 import pdfminer.utils
 
 from .types import Page, Outline, AnnotationType, Annotation, Document
-from .utils import cleanup_text, decode_datetime, format_alpha, format_roman
+from .utils import cleanup_text, decode_datetime
 
 pdfminer.settings.STRICT = False
 
-logger = logging.getLogger(__name__)
+logger = logging.getLogger('pdfannots')
 
-ANNOT_SUBTYPES: typing.Dict[PSLiteral, AnnotationType] = {
+ANNOT_SUBTYPES: typ.Dict[PSLiteral, AnnotationType] = {
     PSLiteralTable.intern(e.name): e for e in AnnotationType}
+"""Mapping from PSliteral to our own enumerant, for supported annotation types."""
+
+IGNORED_ANNOT_SUBTYPES = \
+    frozenset(PSLiteralTable.intern(n) for n in (
+        'Link',   # Links are used for internal document links (e.g. to other pages).
+        'Popup',  # Controls the on-screen appearance of other annotations. TODO: we may want to
+                  # check for an optional 'Contents' field for alternative human-readable contents.
+    ))
+"""Annotation types that we ignore without issuing a warning."""
 
 
 def _mkannotation(
-    pa: typing.Dict[str, typing.Any],
+    pa: typ.Dict[str, typ.Any],
     page: Page
-) -> typing.Optional[Annotation]:
+) -> typ.Optional[Annotation]:
     """
     Given a PDF annotation, capture relevant fields and construct an Annotation object.
 
@@ -46,15 +55,15 @@ def _mkannotation(
 
     subtype = pa.get('Subtype')
     annot_type = None
-    if isinstance(subtype, PSLiteral):
-        try:
-            annot_type = ANNOT_SUBTYPES[subtype]
-        except KeyError:
-            pass
+    assert isinstance(subtype, PSLiteral)
+    try:
+        annot_type = ANNOT_SUBTYPES[subtype]
+    except KeyError:
+        pass
 
     if annot_type is None:
-        if subtype is not PSLiteralTable.intern('Link'):
-            logger.warning("Unsupported annotation subtype: %r", subtype)
+        if subtype not in IGNORED_ANNOT_SUBTYPES:
+            logger.warning("Unsupported %s annotation ignored on %s", subtype.name, page)
         return None
 
     contents = pa.get('Contents')
@@ -62,9 +71,13 @@ def _mkannotation(
         # decode as string, normalise line endings, replace special characters
         contents = cleanup_text(pdfminer.utils.decode_text(contents))
 
-    coords = pdftypes.resolve1(pa.get('QuadPoints'))
+    # Rect defines the location of the annotation on the page
     rect = pdftypes.resolve1(pa.get('Rect'))
 
+    # QuadPoints are defined only for "markup" annotations (Highlight, Underline, StrikeOut,
+    # Squiggly), where they specify the quadrilaterals (boxes) covered by the annotation.
+    quadpoints = pdftypes.resolve1(pa.get('QuadPoints'))
+
     author = pdftypes.resolve1(pa.get('T'))
     if author is not None:
         author = pdfminer.utils.decode_text(author)
@@ -80,14 +93,16 @@ def _mkannotation(
         createds = pdfminer.utils.decode_text(createds)
         created = decode_datetime(createds)
 
-    return Annotation(page, annot_type, coords, rect,
+    return Annotation(page, annot_type, quadpoints, rect,
                       contents, author=author, created=created)
 
 
-def _get_outlines(doc: PDFDocument) -> typing.Iterator[Outline]:
+def _get_outlines(doc: PDFDocument) -> typ.Iterator[Outline]:
     """Retrieve a list of (unresolved) Outline objects for all recognised outlines in the PDF."""
 
-    def _resolve_dest(dest: typing.Any) -> typing.Any:
+    def _resolve_dest(dest: typ.Any) -> typ.Any:
+        if isinstance(dest, pdftypes.PDFObjRef):
+            dest = pdftypes.resolve1(dest)
         if isinstance(dest, bytes):
             dest = pdftypes.resolve1(doc.get_dest(dest))
         elif isinstance(dest, PSLiteral):
@@ -111,91 +126,22 @@ def _get_outlines(doc: PDFDocument) -> typing.Iterator[Outline]:
         if dest[1] is PSLiteralTable.intern('XYZ'):
             (pageref, _, targetx, targety) = dest[:4]
 
-            if isinstance(pageref, (int, pdftypes.PDFObjRef)):
-                yield Outline(title, pageref, (targetx, targety))
-            else:
+            if not isinstance(pageref, (int, pdftypes.PDFObjRef)):
                 logger.warning("Unsupported pageref in outline: %s", pageref)
-
-
-class PDFNoPageLabels(Exception):
-    pass
-
-
-def _get_page_labels(doc: PDFDocument) -> typing.Iterator[str]:
-    """
-    Generate page label strings for the PDF document.
-
-    If the document includes page labels, return a generator of strings, one per page.
-    If not, raise PDFNoPageLabels.
-    """
-    assert doc.catalog is not None
-
-    try:
-        labels_tree = pdftypes.dict_value(doc.catalog['PageLabels'])
-    except (pdftypes.PDFTypeError, KeyError):
-        raise PDFNoPageLabels
-
-    total_pages = pdftypes.int_value(pdftypes.dict_value(doc.catalog['Pages'])['Count'])
-
-    def walk_number_tree(td: typing.Dict[typing.Any, typing.Any]
-                         ) -> typing.Iterator[
-                             typing.Tuple[int, typing.Dict[typing.Any, typing.Any]]]:
-        """
-        Walk a number tree node dictionary, yielding (page index, label dict) tuples.
-
-        See PDF spec, section 3.8.5.
-        """
-        if 'Nums' in td:  # Leaf node
-            objs = pdftypes.list_value(td['Nums'])
-            for (k, v) in pdfminer.utils.choplist(2, objs):
-                yield pdftypes.int_value(k), pdftypes.dict_value(v)
-
-        if 'Kids' in td:  # Intermediate node
-            for child_ref in pdftypes.list_value(td['Kids']):
-                yield from walk_number_tree(pdftypes.dict_value(child_ref))
-
-    # Pass 1: find index ranges
-    range_indices: typing.List[int] = []
-    label_dicts: typing.List[typing.Dict[typing.Any, typing.Any]] = []
-    for (index, d) in walk_number_tree(labels_tree):
-        assert 0 <= index < total_pages
-        if range_indices == []:
-            assert index == 0  # Tree must include page index 0
-        else:
-            assert index > range_indices[-1]  # Tree must be sorted
-
-        range_indices.append(index)
-        label_dicts.append(d)
-
-    # Pass 2: emit page labels
-    for i in range(len(range_indices)):
-        range_start = range_indices[i]
-        range_limit = range_indices[i + 1] if i + 1 < len(range_indices) else total_pages
-
-        d = label_dicts[i]
-        style = d.get('S')
-        prefix = pdfminer.utils.decode_text(pdftypes.str_value(d.get('P', b'')))
-        first = pdftypes.int_value(d.get('St', 1))
-
-        for value in range(first, first + range_limit - range_start):
-            if style is PSLiteralTable.intern('D'):    # Decimal arabic numerals
-                label = str(value)
-            elif style is PSLiteralTable.intern('R'):  # Uppercase roman numerals
-                label = format_roman(value).upper()
-            elif style is PSLiteralTable.intern('r'):  # Lowercase roman numerals
-                label = format_roman(value)
-            elif style is PSLiteralTable.intern('A'):  # Uppercase letters A-Z, AA-ZZ, etc.
-                label = format_alpha(value).upper()
-            elif style is PSLiteralTable.intern('a'):  # Lowercase letters a-z, aa-zz, etc.
-                label = format_alpha(value)
             else:
-                label = ''
+                if targetx is None or targety is None:
+                    # Treat as a general reference to the page
+                    target = None
+                else:
+                    target = (targetx, targety)
+                    if not all(isinstance(v, (int, float)) for v in target):
+                        logger.warning("Unsupported target in outline: (%r, %r)", targetx, targety)
+                        target = None
 
-            yield prefix + label
+                yield Outline(title, pageref, target)
 
 
-class _PDFProcessor(PDFLayoutAnalyzer):  # type:ignore
-    # (pdfminer lacks type annotations)
+class _PDFProcessor(PDFLayoutAnalyzer):
     """
     PDF processor class.
 
@@ -207,17 +153,17 @@ class _PDFProcessor(PDFLayoutAnalyzer):  # type:ignore
     CONTEXT_CHARS = 256
     """Maximum number of recent characters to keep as context."""
 
-    page: typing.Optional[Page]     # Page being processed.
-    charseq: int                    # Character sequence number within the page.
-    lineseq: int                    # Line sequence number within the page.
-    recent_text: typing.Deque[str]  # Rotating buffer of recent text, for context.
-    _lasthit: typing.FrozenSet[Annotation]  # Annotations hit by the most recent character.
-    _curline: typing.Set[Annotation]        # Annotations hit somewhere on the current line.
+    page: typ.Optional[Page]                # Page being processed.
+    charseq: int                            # Character sequence number within the page.
+    compseq: int                            # Component sequence number within the page.
+    recent_text: typ.Deque[str]             # Rotating buffer of recent text, for context.
+    _lasthit: typ.FrozenSet[Annotation]     # Annotations hit by the most recent character.
+    _curline: typ.Set[Annotation]           # Annotations hit somewhere on the current line.
 
     # Stores annotations that are subscribed to receive their post-annotation
     # context. The first element of each tuple, on which the list is sorted, is
     # the sequence number of the last character to hit the annotation.
-    context_subscribers: typing.List[typing.Tuple[int, Annotation]]
+    context_subscribers: typ.List[typ.Tuple[int, Annotation]]
 
     def __init__(self, rsrcmgr: PDFResourceManager, laparams: LAParams):
         super().__init__(rsrcmgr, laparams=laparams)
@@ -229,7 +175,7 @@ class _PDFProcessor(PDFLayoutAnalyzer):  # type:ignore
     def clear(self) -> None:
         """Reset our internal per-page state."""
         self.charseq = 0
-        self.lineseq = 0
+        self.compseq = 0
         self.recent_text.clear()
         self.context_subscribers.clear()
         self._lasthit = frozenset()
@@ -257,13 +203,13 @@ class _PDFProcessor(PDFLayoutAnalyzer):  # type:ignore
 
         self.page = None
 
-    def update_lineseq(self, line: LTTextLine) -> None:
+    def update_pageseq(self, component: LTComponent) -> None:
         """Assign sequence numbers for objects on the page based on the nearest line of text."""
         assert self.page is not None
-        self.lineseq += 1
+        self.compseq += 1
 
         for x in itertools.chain(self.page.annots, self.page.outlines):
-            x.update_pageseq(line, self.lineseq)
+            x.update_pageseq(component, self.compseq)
 
     def test_boxes(self, item: LTComponent) -> None:
         """Update the set of annotations whose boxes intersect with the area of the given item."""
@@ -340,9 +286,10 @@ class _PDFProcessor(PDFLayoutAnalyzer):  # type:ignore
 
         Ref: https://pdfminersix.readthedocs.io/en/latest/topic/converting_pdf_to_text.html
         """
-        # Assign sequence numbers to items on the page based on their proximity to lines of text.
-        if isinstance(item, LTTextLine):
-            self.update_lineseq(item)
+        # Assign sequence numbers to items on the page based on their proximity to lines of text or
+        # to figures (which may contain bare LTChar elements).
+        if isinstance(item, (LTTextLine, LTFigure)):
+            self.update_pageseq(item)
 
         # If it's a container, recurse on nested items.
         if isinstance(item, LTContainer):
@@ -373,10 +320,10 @@ class _PDFProcessor(PDFLayoutAnalyzer):  # type:ignore
 
 
 def process_file(
-    file: typing.BinaryIO,
+    file: typ.BinaryIO,
     *,  # Subsequent arguments are keyword-only
-    columns_per_page: typing.Optional[int] = None,
-    emit_progress_to: typing.Optional[typing.TextIO] = None,
+    columns_per_page: typ.Optional[int] = None,
+    emit_progress_to: typ.Optional[typ.TextIO] = None,
     laparams: LAParams = LAParams()
 ) -> Document:
     """
@@ -403,12 +350,12 @@ def process_file(
 
     emit_progress(file.name)
 
-    # Step 1: retrieve outlines if present. Each outline refers to a page, using
+    # Retrieve outlines if present. Each outline refers to a page, using
     # *either* a PDF object ID or an integer page number. These references will
     # be resolved below while rendering pages -- for now we insert them into one
     # of two dicts for later.
-    outlines_by_pageno = collections.defaultdict(list)
-    outlines_by_objid = collections.defaultdict(list)
+    outlines_by_pageno: typ.Dict[object, typ.List[Outline]] = collections.defaultdict(list)
+    outlines_by_objid: typ.Dict[object, typ.List[Outline]] = collections.defaultdict(list)
 
     try:
         for o in _get_outlines(doc):
@@ -421,27 +368,14 @@ def process_file(
     except Exception as ex:
         logger.warning("Failed to retrieve outlines: %s", ex)
 
-    # Step 2: retrieve page labels, if present.
-    page_labels: typing.Optional[typing.Iterator[str]] = _get_page_labels(doc)
-
-    # Step 3: iterate over all the pages, constructing page objects.
+    # Iterate over all the pages, constructing page objects.
     result = Document()
     for (pageno, pdfpage) in enumerate(PDFPage.create_pages(doc)):
         emit_progress(" %d" % (pageno + 1))
 
-        page = Page(pageno, pdfpage.pageid, pdfpage.mediabox, columns_per_page)
+        page = Page(pageno, pdfpage.pageid, pdfpage.label, pdfpage.mediabox, columns_per_page)
         result.pages.append(page)
 
-        # Retrieve the page's label, but stop trying after an exception is raised.
-        if page_labels is not None:
-            try:
-                page.label = page_labels.__next__()
-            except PDFNoPageLabels:
-                page_labels = None
-            except Exception as ex:
-                logger.warning("Failed to parse page labels: %s", ex)
-                page_labels = None
-
         # Resolve any outlines referring to this page, and link them to the page.
         # Note that outlines may refer to the page number or ID.
         for o in (outlines_by_objid.pop(page.objid, [])
@@ -452,9 +386,11 @@ def process_file(
         # Construct Annotation objects, and append them to the page.
         for pa in pdftypes.resolve1(pdfpage.annots) if pdfpage.annots else []:
             if isinstance(pa, pdftypes.PDFObjRef):
-                annot = _mkannotation(pdftypes.dict_value(pa), page)
-                if annot is not None:
-                    page.annots.append(annot)
+                annot_dict = pdftypes.dict_value(pa)
+                if annot_dict:  # Would be empty if pa is a broken ref
+                    annot = _mkannotation(annot_dict, page)
+                    if annot is not None:
+                        page.annots.append(annot)
             else:
                 logger.warning("Unknown annotation: %s", pa)
 
@@ -472,6 +408,10 @@ def process_file(
         page.annots.sort()
         page.outlines.sort()
 
+        # Give the annotations a chance to update their internals
+        for a in page.annots:
+            a.postprocess()
+
     emit_progress("\n")
 
     device.close()
